/*
 *	 Capricorn.h
 *
 *	 Copyright 2004 Olivier De Smet
 */

//
//   header file for capricorn processor and chipsets
//

#define	SWORD SHORT							// signed   16 Bit variable

typedef struct
{
	WORD	iopst;						// IO processor state
	BYTE	ccr;						// ccr register						
	BYTE	psr;						// psr register
	BYTE	ib;							// ib register
	BYTE	ob;							// ob register
	DWORD	ioptimer;					// timer iop   (internal for waiting)
	BYTE	streg[16];					// iop status registers (max 16 ?) RO
	BYTE	ctreg[32];					// iop control register (max 32 ?) WO
	BYTE	axreg[16];					// iop auxilliary register (max 16 ?) (for HP-IB) RW
	BYTE	iop;						// io processor present (!= 0 , then type cf SR0)
	VOID	**iop_ifc;					// interface control table functions
	VOID	*iop_input;					// input command function
	VOID	*iop_output;				// output command function
	VOID	*iop_send;					// send command function			
	VOID	*iop_reset;					// reset function (depend on interface)
	BOOL	intiop;						// interrupt
	BOOL	ccrwr;						// calculator control register written (usefull ?)
	BYTE	dt;							// var for iop (internal)

} IOP;

typedef struct
{
	SWORD	nPosX;							// position of window
	SWORD	nPosY;
#if defined HP85
	SWORD	nPPosX;							// position of printer window
	SWORD	nPPosY;
#endif
	BYTE	type;							// computer type

	BYTE	ARP;
	BYTE	DRP;
	BYTE	R[64];							// registers R00-R77 (octal)
	BYTE	E;								// E register (1 nibble)
	BOOL	DCM;							// DCM flag
	BOOL	CY;								// CY flag
	BOOL	OVF;							// OVF flag
	BOOL	LSB;							// LSB flag
	BOOL	MSB;							// MSB flag
	BOOL	Z;								// Z flag
	BOOL	LDZ;							// LDZ flag
	BOOL	RDZ;							// RDZ flag

	DWORD	cycles;							// oscillator cycles
	DWORD	cycles_reserved;				// Capricorn instructions count
	WORD	dcycles;						// last instruction duration

	LPBYTE	Rom;							// system rom (0x0000 - 0x5FFF)
	BOOL	RomPage[256];					// paged roms online
	BYTE	Rom3;							// current paged rom
	WORD	RomCrc;							// crc for check state file
	BYTE	IORam[0x200];					// IO control registers (not currently used)
	LPBYTE	Ram;							// pointer to ram							
	DWORD	RamSize;						// ram size (standard & EMS) 
	DWORD	Ptr1;							// EMC pointer 1
	DWORD	Ptr2;							// EMC pointer 2

	LPBYTE	Video;							// Video Ram
	DWORD	VideoSize;						// HP 87 always 16 Kb of video, now can be 64Kb also
#if defined HP87
	BYTE	crtsts;							/* IO crt status HP87
											   Bit  Write 1			/	0			Read  1			/	0
											   0    Read request	/				Data ready		/ Data not ready
											   1    Blank			/ Unblank		Blank			/ Unblank
											   2	Power-down		/ Power-up		Power-down		/ Power-up
											   3	Page 24			/ Page 16		24 lines		/ 16 lines
											   4									Retrace time	/ Display time
											   5	Inverse			/ Noninverse	Inverse			/ Noninverse
											   6	All				/ Normal		All				/ Normal
											   7    Graphics		/ Alpha			Graphics		/ Alpha
											*/
#endif
#if defined HP85
	BYTE	crtrsts;
	BYTE	crtwsts;						/*   HP 85 always 8 Kb of video, accessed as nibbles
											   IO crt status HP85
											   Bit  Write 1			/	0			Read  1			/	0
											   0    Read request	/				Data ready		/ Data not ready
											   1    Wipeout			/ Unwipe		Display time	/ Retrace time		
											   2	Power-Down		/ Power Up		
											   3	     			
											   4									
											   5									always 1 
											   6									always 1
											   7    Graphics		/ Alpha			Busy			/
											*/
#endif

	WORD	crtsad;							// IO crt start address (WO)
	WORD	crtoldsad;						// previous crt start address
	WORD	crtbad;							// IO crt byte address (WO) auto inc
	DWORD	crttimer;						// timer for retrace time (to count 1/60 sec)
#if defined HP85
	WORD	crtcount;						// line counter for service rom
#endif 
	BYTE	crtold;							// previous crt status written							

#if defined HP85
	BYTE	prtsts;
	BYTE	prtdat;
	WORD	prtptr;
#endif

	BYTE	keycode;						// IO keycode
	BYTE	keysts;							// IO keyboard status
	BOOL	keyenable;						// keyboard enable
	BOOL	keyscanner;						// keyboard scanner enable
	BOOL	intkbd;							// keyboard interrupt
	BOOL	KbdShift;						// shift internal kbd status
	BOOL	KbdCtrl;						// ctrl internal kdb status (not used)
	BOOL	KbdCaps;						// caps internal kbd status (not used)
	BOOL	KbdDown;						

	DWORD	ptimer[4];						// preset count
	DWORD	timer[4];						// current count
	WORD	dtimer[4];						// counting cycles (613 for 1ms)
	BOOL	etimer[4];						// enable timer
	BOOL	ctimer[4];						// counting timer
	BOOL	inttimer[4];					// interrupt esked		
	BYTE	clksts;							// IO timer mode
	BYTE	timern;							// internal addressed timer
	BOOL    clkaccess;						// IO write on CLKDAT

	BYTE	intrsc;							// IO info byte
	IOP		iop[11];						// IO processor structure 

	DWORD	annun;							// bitmap for 32 annunciators
	TCHAR	D70x_disk[10][MAX_PATH];		// name of 9121 image disc
	BYTE    D70x_type;						// type of discs (mini, 8", HD)
	BOOL	D70x_ro[10];					// ro ?

	BOOL	inte;							// global interrupt enable flag
	BOOL	idle;							// pc passed by 0x003A with R16 and R17 null
											// basic EXEC idle loop
} CHIPSET;

