/*
 *   Disasm.c
 * 
 *   Copyright 2004 Olivier De Smet 
 */

//
//   Capricorn CPU disassembler
//

#include "StdAfx.h"
#include "HP8x.h"
#include "mops.h"

#define TAB_SKIP	8

static BYTE dARP = 0xFF;		// current ARP for disasm
static BYTE dDRP = 0xFF;		// current ARP for disasm

static BYTE lARP[256];	// stack of ARP opcodes to be used by opcode
static BYTE pARP;		// pointer on this stack
static BYTE lDRP[256];	// stack of DRP opcodes to be used by opcode
static BYTE pDRP;		// pointer on this stack

static LPCTSTR hex = _T("0123456789ABCDEF");

static WORD F_l[64] = {2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,		// field length per register
					   2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,
					   8,7,6,5,4,3,2,1,8,7,6,5,4,3,2,1,
					   8,7,6,5,4,3,2,1,8,7,6,5,4,3,2,1};

#define fl F_l[dDRP]

static LPCTSTR opcode_8[16] =
{
	_T("ELB "), _T("ELM "), _T("ERB "), _T("ERM "),
	_T("LLB "), _T("LLM "), _T("LRB "), _T("LRM "),
	_T("ICB "), _T("ICM "), _T("DCB "), _T("DCM "),
	_T("TCB "), _T("TCM "), _T("NCB "), _T("NCM ")
};

static LPCTSTR opcode_90[8] =
{
	_T("TSB "), _T("TSM "), _T("CLB "), _T("CLM "),
	_T("ORB "), _T("ORM "), _T("XRB "), _T("XRM ")
};

static LPCTSTR opcode_98[8] =
{
	_T("BIN"), _T("BCD"), _T("SAD"), _T("DCE"),
	_T("ICE"), _T("CLE"), _T("RTN"), _T("PAD")
};

static LPCTSTR opcode_A[16] =
{
	_T("LDB "), _T("LDM "), _T("STB "), _T("STM "),
	_T("LDBD "), _T("LDMD "), _T("STBD "), _T("STMD "),
	_T("LDB "), _T("LDM "), _T("STB "), _T("STM "),
	_T("LDBI "), _T("LDMI "), _T("STBI "), _T("STMI ")
};

static LPCTSTR opcode_B[16] =
{
	_T("LDBD "), _T("LDMD "), _T("STBD "), _T("STMD "),
	_T("LDBD "), _T("LDMD "), _T("STBD "), _T("STMD "),
	_T("LDBI "), _T("LDMI "), _T("STBI "), _T("STMI "),
	_T("LDBI "), _T("LDMI "), _T("STBI "), _T("STMI ")
};

static LPCTSTR opcode_C[16] =
{
	_T("CMB "), _T("CMM "), _T("ADB "), _T("ADM "),
	_T("SBB "), _T("SBM "), _T("JSB "), _T("ANM "),
	_T("CMB "), _T("CMM "), _T("ADB "), _T("ADM "),
	_T("SBB "), _T("SBM "), _T("JSB "), _T("ANM ")
};

static LPCTSTR opcode_D[16] =
{
	_T("CMBD "), _T("CMMD "), _T("ADBD "), _T("ADMD "),
	_T("SBBD "), _T("SBMD "), _T("???"), _T("ANMD "),
	_T("CMBD "), _T("CMMD "), _T("ADBD "), _T("ADMD "),
	_T("SBBD "), _T("SBMD "), _T("???"), _T("ANMD ")
};

static LPCTSTR opcode_E[16] =
{
	_T("POBD "), _T("POMD "), _T("POBD "), _T("POMD "),
	_T("PUBD "), _T("PUMD "), _T("PUBD "), _T("PUMD "),
	_T("POBI "), _T("POMI "), _T("POBI "), _T("POMI "),
	_T("PUBI "), _T("PUMI "), _T("PUBI "), _T("PUMI ")
};

static LPCTSTR opcode_F[16] =
{
	_T("JMP "), _T("JNO "), _T("JOD "), _T("JEV "),
	_T("JNG "), _T("JPS "), _T("JNZ "), _T("JZR "),
	_T("JEN "), _T("JEZ "), _T("JNC "), _T("JCY "),
	_T("JLZ "), _T("JLN "), _T("JRZ "), _T("JRN ")
};

//################
//#
//#    Low level subroutines
//#
//################

//
// Return a byte and increment pwAddr
//
static BYTE read_byte (WORD *pwAddr)
{
	BYTE byVal= Nread_nio1(*pwAddr);
	
	*pwAddr = ++(*pwAddr) & 0xFFFF;
	return byVal;
}

//
// Return a Word (16 bits) and increment pwAddr
//
static int read_int (WORD *pwAddr, int n)
{
	int i, t;
	
	for (i = 0, t = 0; i < n; i++)
		t |= read_byte (pwAddr) << (i * 8);
	return t;
}

//
// Append a string to szBuf
//
static LPTSTR append_str (LPTSTR szBuf, LPCTSTR szStr)
{
	while ((*szBuf = *szStr++))
		szBuf++;
	return szBuf;
}

//
// Append TAB as space to szBuf
//
static LPTSTR append_tab (LPTSTR szBuf)
{
	int n;
	LPTSTR p;
	
	n = TAB_SKIP - (lstrlen (szBuf) % TAB_SKIP);
	p = &szBuf[lstrlen (szBuf)];
	while (n--)
		*p++ = _T(' ');
	*p = 0;
	return p;
}

//
// append n immediate bytes from pwAddr
//
static LPTSTR append_imm_byte (LPTSTR szBuf, WORD *pwAddr, int n)
{
	int i, j;
	BYTE byVal;
	
	j = 0;
	for (i = 0; i < n; i++)
	{
		byVal = read_byte(pwAddr);
		*szBuf++ = hex[byVal >> 4];
		*szBuf++ = hex[byVal % 16];
		if (i < n - 1)
			*szBuf++ = ',';
	}
	*szBuf = 0;
	return szBuf;
}

//
// append an adress (2 bytes) in hexa
//
static LPTSTR append_addr (LPTSTR szBuf, WORD wAddr)
{
	int shift;
	long mask;
	
	for (mask = 0xf000, shift = 12; mask != 0; mask >>= 4, shift -= 4)
		*szBuf++ = hex[(wAddr & mask) >> shift];
	*szBuf = 0;
	return szBuf;
}

//
// append a relative address from lwAddr (for Jxx opcodes)
//
static LPTSTR append_r_addr (LPTSTR szBuf, WORD *pwAddr, WORD wDisp)
{
	if (wDisp & 0x80)
		wDisp |= ~(0x7F);
	*pwAddr += wDisp;
	
	if (wDisp & 0x80)
	{
		szBuf = append_str(szBuf, _T("-"));
		wDisp = - wDisp - 2;
	}
	else
	{
		szBuf = append_str(szBuf, _T("+"));
		wDisp += 2;
	}
	szBuf = append_addr(szBuf, wDisp);
	return szBuf;
}

//
// append wAddr as comment
//
static LPTSTR append_pc_comment (LPTSTR szBuf, WORD wAddr, BOOL bView)
{
	LPTSTR p = szBuf;
	
	if (bView == VIEW_LONG)					// output of address in remarks
	{
		while (lstrlen (szBuf) < 4 * TAB_SKIP)
			p = append_tab (szBuf);
		
		p = append_str (p, _T("# Address: "));
		p = append_addr (p, wAddr);
	}
	else									// output of address in brackets
	{
		while (*p) ++p;
		p = append_str (p, _T(" ["));
		p = append_addr (p, wAddr);
		p = append_str (p, _T("]"));
	}
	return p;
}

//
//   functions on DRP and ARP stack
//     No test on overflow when pushing (but no crash as the pointer is BYTE on a 256 bytes stack :)
//     not usefull, generally we don't dissable more than 17 lines at once from debugger dialog
//     so we can't have more than 17 ARP or DRP opcodes in thoses stacks together
//     Not true, can crash if we disassemble near a boundary and only get ARP & DRP opcodes :(
//     (never got in this situation)
//

static __inline VOID push_ARP(n)
{
	lARP[pARP++] = n;
}

static __inline VOID push_DRP(n)
{
	lDRP[pDRP++] = n;
}

static __inline BYTE pop_ARP()
{
	if (pARP > 0)
		return lARP[--pARP];	// pop it
	else
		return 0xFF;			// nothing in stack
}

static __inline BYTE pop_DRP()
{
	if (pDRP > 0)
		return lDRP[--pDRP];	// pop it
	else
		return 0xFF;			// nothing in stack
}

//
// append the ARP register from the stack to the opcode as index (X)
//
static LPTSTR append_XRP(LPTSTR szBuf)
{
	LPTSTR p = szBuf;
	BYTE a = pop_ARP();		// get ARP

	_ASSERT((a == dARP) || (a == 0xFF));

	*p++ = _T('X');			// say 'X'
	switch(dARP)
	{
		case 0xFF:				// nothing in stack
			*p++ = _T('#');			// say '#' as current ARP
			break;
		case 1:					// 01
			*p++ = _T('*');			// indirect ARP
			break;
		default:
			*p++ = hex[dARP / 8];	// in octal
			*p++ = hex[dARP % 8];
			break;
	}
	return p;
}

//
// append the ARP register from the stack to the opcode 
//
static LPTSTR append_ARP(LPTSTR szBuf)
{
	LPTSTR p = szBuf;
	BYTE a = pop_ARP();		// get ARP

	_ASSERT((a == dARP) || (a == 0xFF));

	*p++ = _T('R');			// say 'R'
	switch(dARP)
	{
		case 0xFF:				// nothing in stack
			*p++ = _T('#');			// say '#' as current ARP
			break;
		case 1:					// 01
			*p++ = _T('*');			// indirect ARP
			break;
		default:
			*p++ = hex[dARP / 8];	// in octal
			*p++ = hex[dARP % 8];
			break;
	}
	return p;
}

//
// append the DRP register from the stack to the opcode 
//
static LPTSTR append_DRP(LPTSTR szBuf)
{
	LPTSTR p = szBuf;
	BYTE a = pop_DRP();		// get DRP

	_ASSERT((a == dDRP) || (a == 0xFF));

	*p++ = _T('R');			// say 'R'
	switch(dDRP)
	{
		case 0xFF:				// nothing in stack
			*p++ = _T('#');			// say '#' as current ARP
			break;
		case 1:					// 01
			*p++ = _T('*');			// indirect ARP
			break;
		default:
			*p++ = hex[dDRP / 8];	// in octal
			*p++ = hex[dDRP % 8];
			break;
	}
	return p;
}

//################
//#
//#    Public functions
//#
//################

//
// init stream disassembly by with current value of ARP and DRP
//
VOID init_disassemble(BYTE arp, BYTE drp)
{
	dARP = arp;
	dDRP = drp;
}

//
// disassemble one opcode
//   take into account of orphan's ARP and DRP opcodes (emit them alone)
//   an ARP or DRP is orphan when it is seen but not directly used in the
//   following instruction.
//
//   take care of some instruction which breaks the code flow (RTN, JMP, PAD, JSB)
//   to forget the current ARP & DRP. Instructions modifying R4-5 registers
//   aren't tracked (it could be possible)
//
WORD disassemble (WORD addr, LPTSTR out, BOOL view)
{
	BYTE n;
	LPTSTR p = out;
	WORD disp, pc;
	BOOL emit_arp = FALSE;							// first loop, don't emit ARP alone
	BOOL emit_drp = FALSE;							// first loop, don't emit DRP alone
	WORD start = addr;

	pARP = 0;
	pDRP = 0;
	while (p == out)								// leave when something is emitted
	{
		n = read_byte(&addr);							// read opcode
		switch (n >> 4)										// high nibble
		{													// start decoding
			case 0:
			case 1:
			case 2:
			case 3:												// ARP opcode
				if (emit_arp)										// should we emit it ?
				{
					p = append_str(p, _T("ARP R"));						// emit it alone
					if (n != 1)
					{
						*p++ = hex[n >> 3];
						*p++ = hex[n % 8];
					}
					else
						*p++ = _T('*');
				}
				else												// no ?
				{
					push_ARP(n);										// push it on ARP stack
					dARP = n;											// make it the current one
				}
				break;
			case 4:
			case 5:
			case 6:
			case 7:												// DRP opcode
				if (emit_drp)										// should we emit it ?
				{
					n -= 64;											// emit it alone
					p = append_str(p, _T("DRP R"));
					if (n != 1)
					{
						*p++ = hex[n >> 3];
						*p++ = hex[n % 8];
					}
					else
						*p++ = _T('*');
				}
				else												// no
				{
					n -= 64;											// push it on DRP stack
					push_DRP(n);										// make it the current one
					dDRP = n;
				}
				break;
			case 8:												// 8x opcodes
				n -= 8*16;											// correct index (just get x)
				p = append_str(p, opcode_8[n]);						// from table
				p = append_DRP(p);									// append DRP from stack
				break;
			case 9:												// 9x opcodes ...
				n -= 9*16;
				if (n < 8)
				{
					p = append_str(p, opcode_90[n]);
					p = append_DRP(p);
					if (n > 3)
					{
						*p++ = _T(',');
						p = append_ARP(p);
					}
				}
				else
				{
					p = append_str(p, opcode_98[n-8]);
					if ((n == 15) || (n == 14))							// PAD or RTN
					{
						dARP = 0xFF;										// forget current ARP
						dDRP = 0xFF;										// forget current DRP
					}
				}
				break;
			case 10:
				n -= 10*16;
				p = append_str(p, opcode_A[n]);
				p = append_DRP(p);
				*p++ = _T(',');
				if ((n < 8) || (n > 11))
					p = append_ARP(p);
				else
				{
					*p++ = _T('=');
					if ((n == 8) || (n == 10))
						p = append_imm_byte(p, &addr, 1);
					else
					{
						if ((dDRP == 0xFF) || (dDRP == 1))
						{
							disp = read_int(&addr, 2);
							p = append_addr(p, disp);
						}
						else
						{
							if (fl == 2)
							{
								disp = read_int(&addr, 2);
								p = append_addr(p, disp);
							}
							else
								p = append_imm_byte(p, &addr, fl);
						}
					}
				}
				break;
			case 11:
				n -= 11*16;
				p = append_str(p, opcode_B[n]);
				p = append_DRP(p);
				*p++ = _T(',');
				if (((n/4) % 2) == 0)
				{
					*p++ = _T('=');
					disp = read_int(&addr, 2);
					p = append_addr(p, disp);
				}
				else
				{
					p = append_XRP(p);
					*p++ = _T(',');
					disp = read_int(&addr, 2);
					p = append_addr(p, disp);
				}
				break;
			case 12:
				n -= 12*16;
				p = append_str(p, opcode_C[n]);
				if (n == 6)										// JSB X,n
				{
					p = append_XRP(p);
					*p++ = _T(',');
					disp = read_int(&addr, 2);
					p = append_addr(p, disp);
					dARP = 0xFF;									// forget current ARP
					dDRP = 0xFF;									// forget current DRP
					break;
				}
				if (n == 14)									// JSB =adr
				{
					*p++ = _T('=');
					disp = read_int(&addr, 2);
					p = append_addr(p, disp);
					dARP = 0xFF;									// forget current ARP
					dDRP = 0xFF;									// forget current DRP
					break;
				}
				p = append_DRP(p);
				*p++ = _T(',');
				if (n < 8)
					p = append_ARP(p);
				else
				{
					*p++ = _T('=');
					if ((n == 8) || (n == 10) || (n == 12))
						p = append_imm_byte(p, &addr, 1);
					else
					{
						if ((dDRP == 0xFF) || (dDRP == 1))
						{
							disp = read_int(&addr, 2);
							p = append_addr(p, disp);
						}
						else
						{
							if (fl == 2)
							{
								disp = read_int(&addr, 2);
								p = append_addr(p, disp);
							}
							else
								p = append_imm_byte(p, &addr, fl);
						}
					}
				}
				break;
			case 13:
				n -= 13*16;
				p = append_str(p, opcode_D[n]);
				if ((n == 6) || (n == 14))
					break;
				p = append_DRP(p);
				*p++ = _T(',');
				if (n < 8)
				{
					*p++ = _T('=');
					disp = read_int(&addr, 2);
					p = append_addr(p, disp);
				}
				else
					p = append_ARP(p);
				break;
			case 14:
				n -= 14*16;
				p = append_str(p, opcode_E[n]);
				p = append_DRP(p);
				*p++ = _T(',');
				if (((n/2) % 2) == 0)
					*p++ = _T('+');
				else
					*p++ = _T('-');
				p = append_ARP(p);
				break;
			case 15:
				n -= 15*16;
				pc = addr+1;
				disp = read_int (&addr, 1);
				p = append_str(p, opcode_F[n]);
				p = append_tab (out);
				p = append_r_addr (p, &pc, disp);
				p = append_pc_comment (out, pc, view);
				if (n == 0)									// is it JMP instruction ?
				{
					dARP = 0xFF;								// forget current ARP
					dDRP = 0xFF;								// forget current DRP
				}
				break;
		}												// decoding finished
		if (p == out)									// if nothing emitted, just ARP or DRP opcode
		{													// prepare second loop
			if (pARP == 2)									// if already 2 ARP opcodes on stack
				emit_arp = TRUE;								// we should emit at least the first orphan
			if (pDRP == 2)									// if already 2 DRP opcodes on stack
				emit_drp = TRUE;								// we should emit at least the first orphan
			if (emit_arp || emit_drp)						// if we should emit something
				addr = start;									// restart decoding from the beginning
		}														
		else											// we emit something (first or further loop)
			if (!emit_arp && !emit_drp)						// not an orphan ARP or DRP
			{
				if ((pARP !=0) || (pDRP !=0))					// but there are some orphans ARP or DRP
				{
					p = out;									// forget what we emit
					addr = start;								// restart decoding from the beginning
					if (pARP != 0)								// orphan's ARP	?
						emit_arp = TRUE;							// emit at least one on the next loop
					if (pDRP != 0)								// orphan's DRP ?
						emit_drp = TRUE;							// emit at least one on the next loop
				}
			}
		*p = 0;											// close the string with a NULL byte
	}
	return addr;								// done,  return the next address to disassemble
}
