/*
 *   Hp-2225.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   HP2225A printer as ADDRESS 1 simulation
//
//   no docs about protocol, just some trials and error ....
//   dump on a file (you can open it to copy & paste when the emulator run)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "io.h"								// I/O definitions

#define HP2225ADR 0x1						// HP-IB address

//#define DEBUG_HP2225						// debug flag

#if defined DEBUG_HP2225
static WORD HP2225_ADR[] = {				// states of automata
	0, 1000, 1100, 1400, 
	10000
};

static TCHAR *HP2225_TXT[] ={				// labels of states of automata 
	_T("Idle"), _T("MSA received"), _T("MLA received"), _T("MTA received"),
	_T("Receive data")
};

#define NLAB (sizeof(HP2225_ADR)/sizeof(HP2225_ADR[0]))

static TCHAR *HP2225_LAB(WORD d)			// find a label for the state d
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (HP2225_ADR[i] == d)
			return HP2225_TXT[i];
	return NULL;
}
#endif

// hp2225 variables

static WORD st2225 = 0;								// state of hp2225 controller
static BOOL hp2225_talk = FALSE;					// MTA received ?
static BOOL hp2225_listen = FALSE;					// MLA received ?
static TCHAR hp2225_file[MAX_PATH];					// file for printing
static WORD hp2225_fn = 0;							// number for file name
static HANDLE  hp2225_hfile = NULL;					// handle of file
static BYTE hc_2225[0x200];							// stack of commands
static WORD hc_2225_hi = 0;							// hi mark
static WORD hc_2225_lo = 0;							// lo mark
static BYTE	hd_2225[0x200];							// stack of data in
static WORD hd_2225_hi = 0;							// hi mark
static WORD hd_2225_lo = 0;							// lo mark

BOOL hp2225_ppol_e = TRUE;							// parallel poll enabled

//################
//#
//#    Low level subroutines
//#
//################

BOOL hp2225_pop_c(BYTE *c)
{
	if (hc_2225_hi == hc_2225_lo)
		return FALSE;
	*c = hc_2225[hc_2225_lo++];
	hc_2225_lo &= 0x1FF;
	return TRUE;
}

BOOL hp2225_pop_d(BYTE *c)
{
	if (hd_2225_hi == hd_2225_lo)
		return FALSE;
	*c = hd_2225[hd_2225_lo++];
	hd_2225_lo &= 0x1FF;
	return TRUE;
}

BOOL hp2225_push_c(BYTE addr, BYTE c)	// push on stack hc_2225
{
	if (addr != HP2225ADR)
		return FALSE;
	hc_2225[hc_2225_hi++] = c;
	hc_2225_hi &= 0x1FF;
	_ASSERT(hc_2225_hi != hc_2225_lo);
	return TRUE;
}

BOOL hp2225_push_d(BYTE addr, BYTE d)	// push on stack hd_2225
{	
	if (addr != HP2225ADR || !hp2225_listen)
		return FALSE;
	hd_2225[hd_2225_hi++] = d;
	hd_2225_hi &= 0x1FF;
	_ASSERT(hd_2225_hi != hd_2225_lo);
	return TRUE;
}

//################
//#
//#    Public functions
//#
//################

VOID DoHP2225(VOID)
{
#if defined DEBUG_HP2225
WORD hp2225st;
#endif

BYTE c;
int l;

	#if defined DEBUG_HP2225
	hp2225st = st2225;
	#endif

	if (hc_2225_hi != hc_2225_lo)
	{
		hp2225_pop_c(&c);
		c &= 0x7F;						// remove parity
		switch(c)
		{
			case (0x60 + HP2225ADR):		// my secondary address
//				_ASSERT(0);
				break;
			case (0x20 + HP2225ADR):		// my listen address
				hp2225_listen = TRUE;
				st2225 = 10000;
				break;
			case (0x40 + HP2225ADR):		// my talk address
				hp2225_talk = TRUE;
				st2225 = 1400;
				break;
			case 0x3F:						// unlisten
				hp2225_listen = FALSE;
				st2225 = 0;
				break;
			case 0x5F:						// untalk
				hp2225_talk = FALSE;
				st2225 = 0;
				break;
			default:						// other
				break;
		}
	}

	if (st2225)
		switch (st2225)
		{
		// IDLE STATE
			case 0:
				break;
		// receive my talk address
			case 1400:
				_ASSERT(0);
				break;
		// MLA: receive data
			case 10000:
				if (hp2225_listen)
				{
					if (hp2225_pop_d(&c))
					{
						if (hp2225_hfile != NULL)
						{
							WriteFile(hp2225_hfile, &c, 1, &l, NULL);
						}
					}
				}
				break;
			default:
				_ASSERT(0);
				break;
		}
	#if defined DEBUG_HP2225
	{
		TCHAR buffer[256];
		TCHAR *TXT = HP2225_LAB((WORD)(st2225));
		int k;
		unsigned long l;

		if (hp2225st != st2225)
		{
			if (TXT == NULL)
				k = wsprintf(buffer,_T("%04X: HP2225 : %u->%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp2225st, st2225);
			else
				k = wsprintf(buffer,_T("%04X: HP2225 : %u->%u ( %s )\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp2225st, st2225, TXT);
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
			OutputDebugString(buffer);
		}
	}
	#endif
}

BOOL hp2225_save(VOID)
{
	if (hp2225_hfile != NULL)
	{
		CloseHandle(hp2225_hfile);
		hp2225_fn++;
	}
	return TRUE;
}

BOOL hp2225_eject(VOID)
{
	hp2225_save();
	return TRUE;
}

BOOL hp2225_widle(VOID)
{
	DWORD dwRefTime;

	SuspendDebugger();									// suspend debugger

	dwRefTime = timeGetTime();
	Chipset.idle = FALSE;
	Sleep(10);
	// wait for the Iddle state with 5 sec timeout
	while (((timeGetTime() - dwRefTime) < 5000L) && !(st2225 == 0))
		Sleep(10);

	if (st2225 == 0)									// not timeout, hp2225 is idle
		return FALSE;									// idle state
	else
		ResumeDebugger();								// timeout, resume to debugger

	return TRUE;										// hp2225 was busy
}

VOID hp2225_init(VOID)
{
	TCHAR szFilename[MAX_PATH];
	int k;

	st2225 = 0;								// state of hp9121 controller
	hp2225_talk = FALSE;					// MTA received ?
	hp2225_listen = FALSE;					// MLA received ?
	hp2225_ppol_e = TRUE;					// parallel poll enabled
	hc_2225_hi = 0;							// hi mark
    hc_2225_lo = 0;							// lo mark
	hd_2225_hi = 0;							// hi mark
	hd_2225_lo = 0;							// lo mark

	if (hp2225_hfile == NULL)
	{
		k = wsprintf(szFilename, _T("printer-%03d.txt"), hp2225_fn);
		szFilename[k] = 0x00;
		hp2225_hfile = CreateFile(szFilename,
					   GENERIC_WRITE,
					   FILE_SHARE_READ,
					   NULL,
					   CREATE_ALWAYS,
					   FILE_FLAG_SEQUENTIAL_SCAN,
					   NULL);
		if (hp2225_hfile == INVALID_HANDLE_VALUE)
		{
			hp2225_hfile = NULL;
			return;
		}

	}
}

VOID hp2225_stop(VOID)
{
	hp2225_save();
}
