/*
 *   Hp-ib.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   HP-IB controller as ADDRESS 21 
//
//   "Assembly language I/O reference manual" HP-83/85 Number 00085-90818
//   ASM_IO_REF_83_85.pdf from Andr Koppel
//   "Original 320 rom listing from HP"
//   ROM320_P1.pdf from Andr Koppel
//
//   just get and send bytes from stacks :)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "io.h"								// I/O definitions

//#define DEBUG_HPIB

#if defined DEBUG_HPIB
static TCHAR *HPIB_CMD[] ={
	_T(""),_T("GTL"),_T(""),_T(""),_T("SDC"),_T("PPC"),_T(""),_T(""),_T("GET"),_T("TCT"),
	_T(""),_T(""),_T(""),_T(""),_T(""),_T(""),_T(""),_T("LLO"),_T(""),_T(""),
	_T("DCL"),_T("PPU"),_T(""),_T(""),_T("SPE"),_T("SPD"),_T(""),_T(""),_T(""),_T(""),
	_T(""),_T(""),
	_T("MLA0"),_T("MLA1"),_T("MLA2"),_T("MLA3"),_T("MLA4"),_T("MLA5"),_T("MLA6"),_T("MLA7"),_T("MLA8"),_T("MLA9"),
	_T("MLA10"),_T("MLA11"),_T("MLA12"),_T("MLA13"),_T("MLA14"),_T("MLA15"),_T("MLA16"),_T("MLA17"),_T("MLA18"),_T("MLA19"),
	_T("MLA20"),_T("MLA21"),_T("MLA22"),_T("MLA23"),_T("MLA24"),_T("MLA25"),_T("MLA26"),_T("MLA27"),_T("MLA28"),_T("MLA29"),
	_T("MLA30"),_T("UNL"),
	_T("MTA0"),_T("MTA1"),_T("MTA2"),_T("MTA3"),_T("MTA4"),_T("MTA5"),_T("MTA6"),_T("MTA7"),_T("MTA8"),_T("MTA9"),
	_T("MTA10"),_T("MTA11"),_T("MTA12"),_T("MTA13"),_T("MTA14"),_T("MTA15"),_T("MTA16"),_T("MTA17"),_T("MTA18"),_T("MTA19"),
	_T("MTA20"),_T("MTA21"),_T("MTA22"),_T("MTA23"),_T("MTA24"),_T("MTA25"),_T("MTA26"),_T("MTA27"),_T("MTA28"),_T("MTA29"),
	_T("MTA30"),_T("UNT"),
	_T("MSA0"),_T("MSA1"),_T("MSA2"),_T("MSA3"),_T("MSA4"),_T("MSA5"),_T("MSA6"),_T("MSA7"),_T("MSA8"),_T("MSA9"),
	_T("MSA10"),_T("MSA11"),_T("MSA12"),_T("MSA13"),_T("MSA14"),_T("MSA15"),_T("MSA16"),_T("MSA17"),_T("MSA18"),_T("MSA19"),
	_T("MSA20"),_T("MSA21"),_T("MSA22"),_T("MSA23"),_T("MSA24"),_T("MSA25"),_T("MSA26"),_T("MSA27"),_T("MSA28"),_T("MSA29"),
	_T("MSA30"),_T("")
};
#endif

// HPIB bus

static WORD h_out[0x200];					// stack of data out of controler (back to computer)
WORD h_out_hi = 0;				
WORD h_out_lo = 0;

BOOL h_push(BYTE b, BYTE st)				// push on data out and status st, 0x80 : EOI
{
	h_out[h_out_hi++] = (WORD)(b | (st << 8));
	h_out_hi &= 0x1FF;
	_ASSERT(h_out_hi != h_out_lo);
	return TRUE;
}

static BOOL h_pop(BYTE *b, BYTE *st)		// pop on data out
{
	if (h_out_hi == h_out_lo)
		return FALSE;
	*b = (BYTE)(h_out[h_out_lo] & 0x0FF);
	*st = (BYTE)(h_out[h_out_lo++] >> 8);
	h_out_lo &= 0x1FF;
	return TRUE;
}

// component on hp-ib bus for controller

typedef struct
{
	const VOID  *send_d;
	const VOID  *send_c;
	const VOID  *init;
	const VOID  *stop;
	const BYTE  addr;
} HPIB_INST;

//
// define the actual instruments used
//
HPIB_INST hpib_bus[] = {
	hp9121_push_d, hp9121_push_c, hp9121_init, hp9121_stop, 0,
	hp2225_push_d, hp2225_push_c, hp2225_init, hp2225_stop, 1,
#if defined HP7470
	hp7470_push_d, hp7470_push_c, hp7470_init, hp7470_stop, 5,
#endif
	NULL, NULL, NULL, NULL, 0
};


// function on hpib

BOOL hpib_receive(BYTE *b, BYTE *st)
{
	return h_pop(b, st);
}

BOOL hpib_send_d(BYTE d, BYTE st) 
{
	int i = 0;

	while (hpib_bus[i].send_d != NULL)
	{
		((BOOL (*)(BYTE, BYTE))hpib_bus[i].send_d)(hpib_bus[i].addr, d);
		i++;
	}
	return TRUE;
}

BOOL hpib_send_c(BYTE c) 
{
	int i = 0;

	#if defined DEBUG_HPIB
	{
		TCHAR buffer[256];
		int k;
		unsigned long l;

		k = wsprintf(buffer,_T("%04X: HPIB : send c :%02X (%s)\n"), Chipset.R[4]|(Chipset.R[5]<<8), c, HPIB_CMD[c & 0x7F]);
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
	#endif

	while (hpib_bus[i].send_c != NULL)
	{
		((BOOL (*)(BYTE, BYTE))hpib_bus[i].send_c)(hpib_bus[i].addr, c);
		i++;
	}
	return TRUE;
}

BOOL hpib_reset(IOP *i)
{
	int j = 0;

	i->ctreg[0] = 0x00;		// CR0
	i->ctreg[1] = 0x00;		// CR1
	i->ctreg[2] = 0x00;		// CR2
	i->ctreg[3] = 0x02;		// CR3
	i->ctreg[16] = 0x02;	// CR16
	i->ctreg[17] = 0x0D;	// CR17
	i->ctreg[18] = 0x0A;	// CR18
	i->ctreg[19] = 0x00;	// CR19
	i->ctreg[20] = 0x00;	// CR20
	i->ctreg[21] = 0x00;	// CR21
	i->ctreg[22] = 0x00;	// CR22
	i->ctreg[23] = 0x00;	// CR23

	i->streg[0] = 0x01;		// SR0
	i->streg[1] = 0x00;		// SR1
	i->streg[2] = 0x40;		// SR2
	i->streg[3] = 0x00;		// SR3
	i->streg[4] = 0x35;		// SR4
	i->streg[5] = 0xA0;		// SR5
	i->streg[6] = 0x00;		// SR6

	while (hpib_bus[j].init != NULL)
	{
		((BOOL (*)(VOID))hpib_bus[j].init)();
		j++;
	}
	return TRUE;
}

VOID hpib_stop(IOP *i)
{
	int j = 0;
	while (hpib_bus[j].stop != NULL)
	{
		((BOOL (*)(VOID))hpib_bus[j].stop)();
		j++;
	}
}

static VOID ifc0(IOP *i)		// ABORT I/O
{
	_ASSERT(0);
}

static VOID ifc1(IOP *i)		// REN = 1
{
	_ASSERT(0);
}

static VOID ifc2(IOP *i)		// REN = 0
{
	_ASSERT(0);
}

static VOID ifc3(IOP *i)		// ATN = 0
{
	_ASSERT(0);
}

static VOID ifc4(IOP *i)		// Perform Parallel Pool
{
// quick hack with harcoded interfaces ... need some works

	if (hp9121_ppol_e)
		i->ib = 0x80;
	else
		i->ib = 0x00;
	if (hp2225_ppol_e)
		i->ib |= 0x40;
	else
		i->ib &= ~(0x40);
#if defined HP7470
	if (hp7470_ppol_e)
		i->ib |= 0x04;
	else
		i->ib &= ~(0x04);
#endif
	i->psr |= 0x01;								// set IBF in psr
}

static VOID ifc5(IOP *i)		// Send "MY TALK ADDRESS"
{
	hpib_send_c((BYTE)(0x40 + (i->streg[4] & 0x1F)));
}

static VOID ifc6(IOP *i)		// Send "MY LISTEN ADDRESS"
{
	hpib_send_c((BYTE)(0x20 + (i->streg[4] & 0x1F)));
}

static VOID ifc7(IOP *i)		// Send EOL character sequence
{
	int j = i->ctreg[16] & 0x07;
	int k = 17;

	while(j--)
	{
		if (j == 0 && (i->ctreg[16] & 0x80))
			hpib_send_d(i->ctreg[k++], 0x80);
		else
			hpib_send_d(i->ctreg[k++], 0);
	}
}

static VOID ifc8(IOP *i)		// BREAK I/O
{
	_ASSERT(0);
}

static VOID ifc9(IOP *i)		// RESUME I/O
{
// nothing done (lack of docs)
}

VOID *hpib_ifc[10] = {
	ifc0, ifc1, ifc2,
	ifc3, ifc4, ifc5, 
	ifc6, ifc7, ifc8, 
	ifc9
};
