/*
 *   Settings.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   do all the settings stuff (with registry)
//
//   taken as is from Emu42
//

#include "StdAfx.h"
#include "HP8x.h"

//
// registry handling
//
#if defined HP87
#define REGISTRYKEY "Software\\HP8x"
#endif
#if defined HP85
#define REGISTRYKEY "Software\\HP85"
#endif

#define ReadString(sec, key, dv, v, sv) GetRegistryString(sec, key, dv, v, sv)
#define ReadInt(sec, key, dv)         GetRegistryInt(sec, key, dv)
#define WriteString(sec, key, v)      WriteReg(sec, key, REG_SZ, (BYTE *) v, (lstrlen(v)+1) * sizeof(*v))
#define WriteInt(sec, key, v)         WriteReg(sec, key, REG_DWORD, (BYTE *) &v, sizeof(int))

//################
//#
//#    Low level subroutines
//#
//################

static VOID ReadReg(LPCTSTR lpSubKey, LPCTSTR lpValueName, LPBYTE lpData, DWORD *pdwSize)
{
	TCHAR lpKey[256] = _T(REGISTRYKEY) _T("\\");

	DWORD retCode,dwType;
	HKEY  hKey;

	lstrcat(lpKey, lpSubKey);				// full registry key

	retCode = RegOpenKeyEx(HKEY_CURRENT_USER,
					       lpKey,
						   0,
						   KEY_QUERY_VALUE,
						   &hKey);
	if (retCode == ERROR_SUCCESS) 
	{
		retCode = RegQueryValueEx(hKey, lpValueName, NULL, &dwType, lpData, pdwSize);
		RegCloseKey(hKey);
	}

	if (retCode != ERROR_SUCCESS)			// registry entry not found
		*pdwSize = 0;						// return zero size
} 

static VOID WriteReg(LPCTSTR lpSubKey, LPCTSTR lpValueName, DWORD dwType, CONST BYTE *lpData, DWORD cbData)
{
	TCHAR lpKey[256] = _T(REGISTRYKEY) _T("\\");

	DWORD retCode;
	HKEY  hKey;
	DWORD dwDisposition;

	lstrcat(lpKey, lpSubKey);				// full registry key

	retCode = RegCreateKeyEx(HKEY_CURRENT_USER,
							 lpKey,
							 0,_T(""),
                             REG_OPTION_NON_VOLATILE,
							 KEY_WRITE,
                             NULL,
							 &hKey,
							 &dwDisposition);
	_ASSERT(retCode == ERROR_SUCCESS);

	RegSetValueEx(hKey, lpValueName, 0, dwType, lpData, cbData);
	RegCloseKey(hKey);
} 

static DWORD GetRegistryString(LPCTSTR lpszSection, LPCTSTR lpszEntry, LPCTSTR lpDefault, LPTSTR lpData, DWORD dwSize)
{
	dwSize *= sizeof(*lpData);				// buffer size in bytes
	ReadReg(lpszSection, lpszEntry, (LPBYTE) lpData, &dwSize);
	if (dwSize == 0)
	{
		lstrcpy(lpData, lpDefault);
		dwSize = lstrlen(lpData);
	}
	else
	{
		dwSize = (dwSize / sizeof(*lpData)) - 1;
	}
	return dwSize;
}

static INT GetRegistryInt(LPCTSTR lpszSection, LPCTSTR lpszEntry, INT nDefault)
{
	UINT  nValue;
	DWORD dwSize = sizeof(nValue);

	ReadReg(lpszSection, lpszEntry, (LPBYTE) &nValue, &dwSize);
	return dwSize ? nValue : nDefault;
}

//################
//#
//#    Public functions
//#
//################

VOID ReadSettings(VOID)
{
	// Files
	ReadString(_T("Files"), _T("HP8xDirectory"), szCurrentDirectory, szHP8xDirectory, ARRAYSIZEOF(szHP8xDirectory));
	bAutoSave = ReadInt(_T("Files"), _T("AutoSave"),bAutoSave);
	bAutoSaveOnExit = ReadInt(_T("Files"), _T("AutoSaveOnExit"), bAutoSaveOnExit);
	// KML
	bAlwaysDisplayLog = ReadInt(_T("KML"), _T("AlwaysDisplayLog"), bAlwaysDisplayLog);
	// Emulator
	bRealSpeed = ReadInt( _T("Emulator"), _T("RealSpeed"), bRealSpeed);
	dwCapricornCycles = ReadInt(_T("Emulator"), _T("CapricornCycles"), dwCapricornCycles);
	bSound = ReadInt(_T("Emulator"), _T("Sound"), bSound);
	bSetDate = ReadInt(_T("Emulator"), _T("SetDate"), bSound);
	SetSpeed(bRealSpeed);					// set speed
	CloseWaves();							// set sound
	if (bSound)
		bSound = InitWaves();
}

VOID WriteSettings(VOID)
{
	// Files
	WriteString(_T("Files"), _T("HP8xDirectory"), szHP8xDirectory);
	WriteInt(_T("Files"), _T("AutoSave"), bAutoSave);
	WriteInt(_T("Files"), _T("AutoSaveOnExit"), bAutoSaveOnExit);
	// KML
	WriteInt(_T("KML"), _T("AlwaysDisplayLog"), bAlwaysDisplayLog);
	// Emulator
	WriteInt(_T("Emulator"), _T("RealSpeed"), bRealSpeed);
	WriteInt(_T("Emulator"), _T("CapricornCycles"), dwCapricornCycles);
	WriteInt(_T("Emulator"), _T("Sound"), bSound);
	WriteInt(_T("Emulator"), _T("SetDate"), bSetDate);
}

VOID ReadLastDocument(LPTSTR szFilename, DWORD nSize)
{
	ReadString(_T("Files"),_T("LastDocument"),_T(""),szFilename,nSize);
}

VOID WriteLastDocument(LPCTSTR szFilename)
{
	WriteString(_T("Files"),_T("LastDocument"),szFilename);
}
